/* 
 * This file is part of PLaSK (https://plask.app) by Photonics Group at TUL
 * Copyright (c) 2022 Lodz University of Technology
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */
#ifndef PLASK__GEOMETRY_UTILS_H
#define PLASK__GEOMETRY_UTILS_H

#include "object.hpp"

/** @file
This file contains some utils useful with geometry classes.
*/

namespace plask {

//TODO at the moment, this template is not used, consider to remove it
/**
 * Lazy cache of object bounding box.
 */
template <int dims>
struct PLASK_API BoundingBoxCache {

    typedef typename Primitive<dims>::Box BoundingBoxT;

private:

    /// Connection object with child. It is necessary since disconnectOnChileChanged doesn't work
    boost::signals2::connection connection_with_object;

    /// Cached bounding box
    BoundingBoxT boundingBox;

    GeometryObjectD<dims>* object;

    bool isFresh; ///< @c true only if value in cache is fresh

    void connect();

public:

    /**
     * Get bounding box of connected object. Read it from cache if it's possible or update cache.
     * @return bounding box of connected object
     * @throw Exception if no object is connected
     */
    const BoundingBoxT& operator()();

    /**
     * Refresh bounding box cache. Called by object changed signal.
     * @param evt
     */
    void onObjectChanged(const GeometryObject::Event& evt);

    /**
     * Set object for which bounding box should be cached.
     * @param object object for which bounding box should be cached, can be nullptr to disconnect
     */
    void setObject(GeometryObjectD<dims>* object);

    /**
     * Set object for which bounding box should be cached.
     * @param object object for which bounding box should be cached
     */
    void setObject(GeometryObjectD<dims>& object) {
        setObject(&object);
    }

    /**
     * Set object for which bounding box should be cached.
     * @param object object for which bounding box should be cached
     */
    void setObject(shared_ptr<GeometryObjectD<dims> > object) {
        setObject(*object);
    }

    /**
     * Get object for which bounding box is cached.
     * @return object for which bounding box is cached or @c nullptr if bounding box is not cached, for any object
     */
    GeometryObjectD<dims>* getObject() const {
        return object;
    }

    /**
     * Construct bounding box cache connected with given object
     * @param object object for which bounding box should be cached, can be nullptr (typically you should later call setObject in such cache)
     */
    BoundingBoxCache(GeometryObjectD<dims>* object = 0): object(object), isFresh(false) {
        connect();
    }

    /**
     * Construct bounding box cache connected with given object
     * @param object object for which bounding box should be cached
     */
    BoundingBoxCache(GeometryObjectD<dims>& object): object(&object), isFresh(false) {
        connect();
    }

    /**
     * Construct bounding box cache connected with given object
     * @param object object for which bounding box should be cached
     */
    BoundingBoxCache(shared_ptr< GeometryObjectD<dims> > object): object(object.get()), isFresh(false) {
        connect();
    }

    ~BoundingBoxCache() {
        connection_with_object.disconnect();
    }

};

PLASK_API_EXTERN_TEMPLATE_STRUCT(BoundingBoxCache<2>)
PLASK_API_EXTERN_TEMPLATE_STRUCT(BoundingBoxCache<3>)

}   // namespace plask

#endif // PLASK__GEOMETRY_UTILS_H
