/*
 * This file is part of PLaSK (https://plask.app) by Photonics Group at TUL
 * Copyright (c) 2022 Lodz University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */
#ifndef PLASK__RECTILINEAR3D_H
#define PLASK__RECTILINEAR3D_H

/** @file
This file contains rectilinear mesh for 3D space.
*/

#include <type_traits>

#include "rectangular_common.hpp"
#include "axis1d.hpp"
#include "../utils/interpolation.hpp"


namespace plask {

/**
 * Rectilinear mesh in 3D space.
 *
 * Includes three 1d rectilinear meshes:
 * - axis0
 * - axis1
 * - axis2
 * Represent all points (c0, c1, c2) such that c0 is in axis0, c1 is in axis1, c2 is in axis2.
 */
class PLASK_API RectilinearMesh3D: public RectangularMeshBase3D /*MeshD<3>*/ {

    typedef std::size_t index_ft(const RectilinearMesh3D* mesh, std::size_t c0_index, std::size_t c1_index, std::size_t c2_index);
    typedef std::size_t index012_ft(const RectilinearMesh3D* mesh, std::size_t mesh_index);

    // our own virtual table, changeable at run-time:
    index_ft* index_f;
    index012_ft* index0_f;
    index012_ft* index1_f;
    index012_ft* index2_f;
    const shared_ptr<MeshAxis>* minor_axis;
    const shared_ptr<MeshAxis>* medium_axis;
    const shared_ptr<MeshAxis>* major_axis;

    void onAxisChanged(Event& e);

    void setChangeSignal(const shared_ptr<MeshAxis>& axis) { if (axis) axis->changedConnectMethod(this, &RectilinearMesh3D::onAxisChanged); }
    void unsetChangeSignal(const shared_ptr<MeshAxis>& axis) { if (axis) axis->changedDisconnectMethod(this, &RectilinearMesh3D::onAxisChanged); }

  public:

    /**
     * Represent FEM-like element in Rectilinear.
     */
    class PLASK_API Element {
        const RectilinearMesh3D& mesh;
        std::size_t index0, index1, index2; // probably this form allows to do most operation fastest in average, low indexes of element corner or just element indexes

        public:

        /**
         * Construct element using mesh and element indexes.
         * @param mesh mesh, this element is valid up to time of this mesh life
         * @param index0, index1, index2 axis 0, 1 and 2 indexes of element (equal to low corrner mesh indexes of element)
         */
        Element(const RectilinearMesh3D& mesh, std::size_t index0, std::size_t index1, std::size_t index2): mesh(mesh), index0(index0), index1(index1), index2(index2) {}

        /**
         * Construct element using mesh and element index.
         * @param mesh mesh, this element is valid up to time of this mesh life
         * @param elementIndex index of element
         */
        Element(const RectilinearMesh3D& mesh, std::size_t elementIndex): mesh(mesh) {
            std::size_t v = mesh.getElementMeshLowIndex(elementIndex);
            index0 = mesh.index0(v);
            index1 = mesh.index1(v);
            index2 = mesh.index2(v);
        }

        /// \return long index of the element
        inline std::size_t getIndex0() const { return index0; }

        /// \return tran index of the element
        inline std::size_t getIndex1() const { return index1; }

        /// \return vert index of the element
        inline std::size_t getIndex2() const { return index2; }

        /// \return long index of the back edge of the element
        inline std::size_t getLowerIndex0() const { return index0; }

        /// \return tran index of the left edge of the element
        inline std::size_t getLowerIndex1() const { return index1; }

        /// \return vert index of the bottom edge of the element
        inline std::size_t getLowerIndex2() const { return index2; }

        /// \return long coordinate of the back edge of the element
        inline double getLower0() const { return mesh.axis[0]->at(index0); }

        /// \return tran coordinate of the left edge of the element
        inline double getLower1() const { return mesh.axis[1]->at(index1); }

        /// \return vert coordinate of the bottom edge of the element
        inline double getLower2() const { return mesh.axis[2]->at(index2); }

        /// \return long index of the front edge of the element
        inline std::size_t getUpperIndex0() const { return index0+1; }

        /// \return tran index of the right edge of the element
        inline std::size_t getUpperIndex1() const { return index1+1; }

        /// \return vert index of the top edge of the element
        inline std::size_t getUpperIndex2() const { return index2+1; }

        /// \return long coordinate of the front edge of the element
        inline double getUpper0() const { return mesh.axis[0]->at(getUpperIndex0()); }

        /// \return tran coordinate of the right edge of the element
        inline double getUpper1() const { return mesh.axis[1]->at(getUpperIndex1()); }

        /// \return vert coordinate of the top edge of the element
        inline double getUpper2() const { return mesh.axis[2]->at(getUpperIndex2()); }

        /// \return size of the element in the long direction
        inline double getSize0() const { return getUpper0() - getLower0(); }

        /// \return size of the element in the tran direction
        inline double getSize1() const { return getUpper1() - getLower1(); }

        /// \return size of the element in the vert direction
        inline double getSize2() const { return getUpper2() - getLower2(); }

        /// \return vector indicating size of the element
        inline Vec<3, double> getSize() const { return getUpUpUp() - getLoLoLo(); }

        /// \return position of the middle of the element
        inline Vec<3, double> getMidpoint() const { return mesh.getElementMidpoint(index0, index1, index2); }

        /// @return index of this element
        inline std::size_t getIndex() const { return mesh.getElementIndexFromLowIndex(getLoLoLoIndex()); }

        /// \return index of the lower left back corner of this element
        inline std::size_t getLoLoLoIndex() const { return mesh.index(getLowerIndex0(), getLowerIndex1(), getLowerIndex2()); }

        /// \return index of the lower left front corner of this element
        inline std::size_t getUpLoLoIndex() const { return mesh.index(getUpperIndex0(), getLowerIndex1(), getLowerIndex2()); }

        /// \return index of the lower right back corner of this element
        inline std::size_t getLoUpLoIndex() const { return mesh.index(getLowerIndex0(), getUpperIndex1(), getLowerIndex2()); }

        /// \return index of the lower right front corner of this element
        inline std::size_t getUpUpLoIndex() const { return mesh.index(getUpperIndex0(), getUpperIndex1(), getLowerIndex2()); }

        /// \return index of the upper left back corner of this element
        inline std::size_t getLoLoUpIndex() const { return mesh.index(getLowerIndex0(), getLowerIndex1(), getUpperIndex2()); }

        /// \return index of the upper left front corner of this element
        inline std::size_t getUpLoUpIndex() const { return mesh.index(getUpperIndex0(), getLowerIndex1(), getUpperIndex2()); }

        /// \return index of the upper right back corner of this element
        inline std::size_t getLoUpUpIndex() const { return mesh.index(getLowerIndex0(), getUpperIndex1(), getUpperIndex2()); }

        /// \return index of the upper right front corner of this element
        inline std::size_t getUpUpUpIndex() const { return mesh.index(getUpperIndex0(), getUpperIndex1(), getUpperIndex2()); }

        /// \return position of the lower left back corner of this element
        inline Vec<3, double> getLoLoLo() const { return mesh(getLowerIndex0(), getLowerIndex1(), getLowerIndex2()); }

        /// \return position of the lower left front corner of this element
        inline Vec<3, double> getUpLoLo() const { return mesh(getUpperIndex0(), getLowerIndex1(), getLowerIndex2()); }

        /// \return position of the lower right back corner of this element
        inline Vec<3, double> getLoUpLo() const { return mesh(getLowerIndex0(), getUpperIndex1(), getLowerIndex2()); }

        /// \return position of the lower right front corner of this element
        inline Vec<3, double> getUpUpLo() const { return mesh(getUpperIndex0(), getUpperIndex1(), getLowerIndex2()); }

        /// \return position of the upper left back corner of this element
        inline Vec<3, double> getLoLoUp() const { return mesh(getLowerIndex0(), getLowerIndex1(), getUpperIndex2()); }

        /// \return position of the upper left front corner of this element
        inline Vec<3, double> getUpLoUp() const { return mesh(getUpperIndex0(), getLowerIndex1(), getUpperIndex2()); }

        /// \return position of the upper right back corner of this element
        inline Vec<3, double> getLoUpUp() const { return mesh(getLowerIndex0(), getUpperIndex1(), getUpperIndex2()); }

        /// \return position of the upper right front corner of this element
        inline Vec<3, double> getUpUpUp() const { return mesh(getUpperIndex0(), getUpperIndex1(), getUpperIndex2()); }

        /// \return this element as rectangular box
        inline Box3D toBox() const { return mesh.getElementBox(index0, index1, index2); }

        /**
         * Check if point @p p is included in rectangle represented by @c this element.
         * @param p point to check
         * @return @c true only if @p p is included in @c this
         */
        bool contains(Vec<3, double> p) const { return toBox().contains(p); }

    };

    /**
     * Wrapper to Rectilinear which allow to access to FEM-like elements.
     *
     * It works like read-only, random access container of @ref Element objects.
     */
    struct PLASK_API Elements {

        static inline Element deref(const RectilinearMesh3D& mesh, std::size_t index) { return mesh.getElement(index); }
    public:
        typedef IndexedIterator<const RectilinearMesh3D, Element, deref> const_iterator;
        typedef const_iterator iterator;

        /// Mesh which elements will be accessible by this.
        const RectilinearMesh3D* mesh;

        /**
         * Create wrapper which allow to access to FEM-like elements of given @p mesh.
         * @param mesh mesh which elements will be accessible by this
         */
        Elements(const RectilinearMesh3D* mesh): mesh(mesh) {}

        /**
         * Get @p i-th element.
         * @param i element index
         * @return @p i-th element
         */
        Element operator[](std::size_t i) const { return Element(*mesh, i); }

        /**
         * Get element with indices \p i0, \p i1, and \p i2.
         * \param i0, i1, i2 element index
         * \return element with indices \p i0, \p i1, and \p i2
         */
        Element operator()(std::size_t i0, std::size_t i1, std::size_t i2) const { return Element(*mesh, i0, i1, i2); }

        /**
         * Get number of elements.
         * @return number of elements
         */
        std::size_t size() const { return mesh->getElementsCount(); }

        /// @return iterator referring to the first element
        const_iterator begin() const { return const_iterator(mesh, 0); }

        /// @return iterator referring to the past-the-end element
        const_iterator end() const { return const_iterator(mesh, size()); }

    };

    /**
     * Element mesh class.
     *
     * This class has changed nearest neighnour interpolation, to consider original mesh boundaries
     */
    template <typename BaseMeshT>
    struct PLASK_API ElementMesh: BaseMeshT {

        /// Original mesh
        const BaseMeshT* originalMesh;

        template <typename... Args>
        ElementMesh(const BaseMeshT* originalMesh, Args... args): BaseMeshT(args...), originalMesh(originalMesh) {}

        /**
        * Calculate (using nearest neighbor interpolation) value of data in point using data in points described by this mesh.
        * Consider original mesh boundaries for point selection
        * \param data values of data in points describe by this mesh
        * \param point point in which value should be calculate
        * \return interpolated value in point @p point
        */
        template <typename RandomAccessContainer>
        auto interpolateNearestNeighbor(const RandomAccessContainer& data, const Vec<3>& point, const InterpolationFlags& flags) const
            -> typename std::remove_reference<decltype(data[0])>::type {
            auto p = flags.wrap(point);
            prepareNearestNeighborInterpolationForAxis(*originalMesh->axis[0], flags, p.c0, 0);
            prepareNearestNeighborInterpolationForAxis(*originalMesh->axis[1], flags, p.c1, 1);
            prepareNearestNeighborInterpolationForAxis(*originalMesh->axis[2], flags, p.c2, 2);
            size_t i0 = originalMesh->axis[0]->findUpIndex(p.c0),
                   i1 = originalMesh->axis[0]->findUpIndex(p.c1),
                   i2 = originalMesh->axis[2]->findUpIndex(p.c2);
            if (i0 == originalMesh->axis[0]->size()) --i0;
            if (i0 != 0) --i0;
            if (i1 == originalMesh->axis[1]->size()) --i1;
            if (i1 != 0) --i1;
            if (i2 == originalMesh->axis[2]->size()) --i2;
            if (i2 != 0) --i2;
            return flags.postprocess(point, data[this->index(i0, i1, i2)]);
        }

    protected:

        bool hasSameNodes(const MeshD<3> &to_compare) const override {
            if (const ElementMesh* c = dynamic_cast<const ElementMesh*>(&to_compare))
                if (this->originalMesh == c->originalMesh) return true;
            return BaseMeshT::hasSameNodes(to_compare);
        }
    };

    /// First, second and third coordinates of points in this mesh.
    const shared_ptr<MeshAxis> axis[3];

    /// Accessor to FEM-like elements.
    Elements elements() const { return Elements(this); }
    Elements getElements() const { return elements(); }

    Element element(std::size_t i0, std::size_t i1, std::size_t i2) const { return Element(*this, i0, i1, i2); }
    Element getElement(std::size_t i0, std::size_t i1, std::size_t i2) const { return element(i0, i1, i2); }

    /**
     * Get an element with a given index @p i.
     * @param i index of the element
     * @return the element
     */
    Element element(std::size_t i) const { return Element(*this, i); }

    /**
     * Get an element with a given index @p i.
     * @param i index of the element
     * @return the element
     */
    Element getElement(std::size_t i) const { return element(i); }

    /**
     * Get element as box.
     * @param index0, index1, index2 index of element
     * @return box of elements with given indexes
     */
    Box3D getElementBox(std::size_t index0, std::size_t index1, std::size_t index2) const {
        return Box3D(axis[0]->at(index0), axis[1]->at(index1), axis[2]->at(index2),
                     axis[0]->at(index0+1), axis[1]->at(index1+1), axis[2]->at(index2+1));
    }

    /**
     * Iteration orders:
     * Every other order is proper permutation of indices.
     * They mean ORDER_major,medium,minor, i.e. the last index changes fastest
     * @see setIterationOrder, getIterationOrder, setOptimalIterationOrder
     */
    enum IterationOrder { ORDER_012, ORDER_021, ORDER_102, ORDER_120, ORDER_201, ORDER_210 };

    /**
     * Choose iteration order.
     * @param order iteration order to use
     * @see IterationOrder
     */
    void setIterationOrder(IterationOrder order);

    /**
     * Get iteration order.
     * @return iteration order used by this mesh
     * @see IterationOrder
     */
    IterationOrder getIterationOrder() const;

    /**
     * Get iteration order as an array, e.g. [1, 0, 2] is returned for ORDER_102.
     * @return the array of length 3, pointer to static memory
     */
    const char* getIterationOrderAsArray() const {
        return &"\0\1\2\0\2\1\1\0\2\1\2\0\2\0\1\2\1\0"[getIterationOrder() * 3];
        //       |0 1 2|0 1 2|0 1 2|0 1 2|0 1 2|0 1 2|
    };

    /**
     * Similar to getIterationOrderAsArray but the resulted array has exchanged indexes with values.
     *
     * Resulted array is indexed by axis number (0, 1, 2) and the bigger value means
     * that the index of the axis changes faster when index of the mesh is changed by 1.
     * @return the array which maps axis number to position in iteration order
     */
    const char* getAxisToIterationOrder() const {
        return &"\0\1\2\0\2\1\1\0\2\2\0\1\1\2\0\2\1\0"[getIterationOrder() * 3];
        //       |0 1 2|0 1 2|0 1 2|0 1 2|0 1 2|0 1 2|
    };

    /**
     * Check if axis with index @p axis_index1 changes slower than one with index @p axis_index2 when the mesh is changed by 1.
     * @param array_index1, array_index2 indexes of axes, 0, 1, or 2
     * @return whether axis with index @p axis_index1 changes slower than one with index @p axis_index2
     */
    bool isChangeSlower(std::size_t axis_index1, std::size_t axis_index2) const {
        return getAxisToIterationOrder()[axis_index1] < getAxisToIterationOrder()[axis_index2];
    }

    /**
     * Set iteration order to the shortest axis changes fastest.
     */
    void setOptimalIterationOrder();

    /**
     * Construct mesh which has all axes of type OrderedAxis and all are empty.
     * @param iterationOrder iteration order
     */
    explicit RectilinearMesh3D(IterationOrder iterationOrder = ORDER_012);

    /**
     * Construct mesh with is based on given 1D meshes
     *
     * @param mesh0 mesh for the first coordinate
     * @param mesh1 mesh for the second coordinate
     * @param mesh2 mesh for the third coordinate
     * @param iterationOrder iteration order
     */
    RectilinearMesh3D(shared_ptr<MeshAxis> mesh0, shared_ptr<MeshAxis> mesh1, shared_ptr<MeshAxis> mesh2, IterationOrder iterationOrder = ORDER_012);

    /**
     * Change axes and iteration order of this mesh.
     * @param mesh0 mesh for the first coordinate
     * @param mesh1 mesh for the second coordinate
     * @param mesh2 mesh for the third coordinate
     * @param iterationOrder iteration order
     */
    void reset(shared_ptr<MeshAxis> mesh0, shared_ptr<MeshAxis> mesh1, shared_ptr<MeshAxis> mesh2, IterationOrder iterationOrder = ORDER_012);

    /**
     * Copy constructor.
     * @param src mesh to copy
     * @param clone_axes whether axes of the @p src should be cloned (if true) or shared (if false; default)
     */
    RectilinearMesh3D(const RectilinearMesh3D& src, bool clone_axes = false);

    /**
     * Change axes and iteration order of this mesh to the ones from @p src.
     * @param src mesh to copy
     */
    void reset(const RectilinearMesh3D& src, bool clone_axes = false);

    RectilinearMesh3D& operator=(const RectilinearMesh3D& src) { reset(src, true); return *this; }

    RectilinearMesh3D& operator=(RectilinearMesh3D&& src) { reset(src, false); return *this; }

    ~RectilinearMesh3D();

    /**
     * Change axis.
     * @param axis_nr number of axis to change
     * @param new_val new value for axis
     * @param fireResized whether to call fireResized()
     */
    void setAxis(std::size_t axis_nr, shared_ptr<MeshAxis> new_val, bool fireResized = true);

    const shared_ptr<MeshAxis> getAxis0() const { return axis[0]; }

    void setAxis0(shared_ptr<MeshAxis> a0) { setAxis(0, a0);  }

    const shared_ptr<MeshAxis> getAxis1() const { return axis[1]; }

    void setAxis1(shared_ptr<MeshAxis> a1) { setAxis(1, a1); }

    const shared_ptr<MeshAxis> getAxis2() const { return axis[2]; }

    void setAxis2(shared_ptr<MeshAxis> a2) { setAxis(2, a2); }

    /**
     * Get numbered axis
     * \param n number of axis
     */
    const shared_ptr<MeshAxis>& getAxis(size_t n) const {
        if (n >= 3) throw Exception("bad axis number");
        return axis[n];
    }

    /// \return major (changing slowest) axis
    inline const shared_ptr<MeshAxis> majorAxis() const {
        return *major_axis;
    }

    /// \return index of major axis
    inline std::size_t majorAxisIndex() const {
        if (major_axis == &axis[0]) return 0;
        if (major_axis == &axis[1]) return 1;
        return 2;
    }

    /// \return middle (between major and minor) axis
    inline const shared_ptr<MeshAxis> mediumAxis() const {
        return *medium_axis;
    }

    /// \return index of middle axis
    inline std::size_t mediumAxisIndex() const {
        if (medium_axis == &axis[0]) return 0;
        if (medium_axis == &axis[1]) return 1;
        return 2;
    }

    /// \return minor (changing fastes) axis
    inline const shared_ptr<MeshAxis> minorAxis() const {
        return *minor_axis;
    }

    /// \return index of minor (changing fastes) axis
    inline std::size_t minorAxisIndex() const {
        if (minor_axis == &axis[0]) return 0;
        if (minor_axis == &axis[1]) return 1;
        return 2;
    }

    /**
      * Compare meshes.
      * @param to_compare mesh to compare
      * @return @c true only if this mesh and @p to_compare represents the same sequence of points
      */
    bool operator==(const RectilinearMesh3D& to_compare) const {
        if (empty()) return to_compare.empty();
        // here we know that axes are non-empty
        if (*axis[0] != *to_compare.axis[0] || *axis[1] != *to_compare.axis[1] || *axis[2] != *to_compare.axis[2]) return false;
        // here we know that axes are equal
        return this->getIterationOrder() == to_compare.getIterationOrder() ||
                (unsigned(axis[0]->size() == 1) + unsigned(axis[1]->size() == 1) + unsigned(axis[2]->size() == 1)) >= 2;
    }

    bool operator!=(const RectilinearMesh3D& to_compare) const {
        return !(*this == to_compare);
    }

    /**
     * Get number of points in the mesh.
     * @return number of points in the mesh
     */
    std::size_t size() const override { return axis[0]->size() * axis[1]->size() * axis[2]->size(); }

    /// @return true only if there are no points in mesh
    bool empty() const override { return axis[0]->empty() || axis[1]->empty() || axis[2]->empty(); }

    /**
     * Get point with given mesh indices.
     * @param index0 index of point in axis0
     * @param index1 index of point in axis1
     * @param index2 index of point in axis2
     * @return point with given @p index
     */
    virtual Vec<3, double> at(std::size_t index0, std::size_t index1, std::size_t index2) const = 0;

    /**
     * Get point with given mesh index.
     * @param index index of point, from 0 to size()-1
     * @return point with given @p index
     */
    Vec<3, double> at(std::size_t index) const override {
        return at(index0(index), index1(index), index2(index));
    }

    /**
     * Get point with given mesh index.
     * Points are in order: (c0[0], c1[0], c2[0]), (c0[1], c1[0], c2[0]), ..., (c0[c0.size-1], c1[0], c2[0]), (c0[0], c1[1], c2[0]), ..., (c0[c0.size()-1], c1[c1.size()-1], c2[c2.size()-1])
     * @param index index of point, from 0 to size()-1
     * @return point with given @p index
     */
    inline Vec<3, double> operator[](std::size_t index) const {
        return at(index0(index), index1(index), index2(index));
    }

    /**
     * Get point with given x and y indexes.
     * @param index0 index of point in axis0
     * @param index1 index of point in axis1
     * @param index2 index of point in axis2
     * @return point with given c0, c1 and c2 indexes
     */
    inline Vec<3, double> operator()(std::size_t index0, std::size_t index1, std::size_t index2) const {
        return at(index0, index1, index2);
    }

    /**
     * Get first coordinate of point in center of Elements.
     * @param index0 index of Elements (axis0 index)
     * @return first coordinate of point point in center of Elements with given index
     */
    double getElementMidpoint0(std::size_t index0) const { return 0.5 * (axis[0]->at(index0) + axis[0]->at(index0+1)); }

    /**
     * Get second coordinate of point in center of Elements.
     * @param index1 index of Elements (axis1 index)
     * @return second coordinate of point point in center of Elements with given index
     */
    double getElementMidpoint1(std::size_t index1) const { return 0.5 * (axis[1]->at(index1) + axis[1]->at(index1+1)); }

    /**
     * Get second coordinate of point in center of Elements.
     * @param index2 index of Elements (axis2 index)
     * @return second coordinate of point point in center of Elements with given index
     */
    double getElementMidpoint2(std::size_t index2) const { return 0.5 * (axis[2]->at(index2) + axis[2]->at(index2+1)); }

    /**
     * Get point in center of Elements.
     * @param index0, index1, index2 index of Elements
     * @return point in center of element with given index
     */
    virtual Vec<3, double> getElementMidpoint(std::size_t index0, std::size_t index1, std::size_t index2) const = 0;

    /**
     * Calculate this mesh index using indexes of c0, c1 and c2.
     * @param c0_index index of c0, from 0 to c0.size()-1
     * @param c1_index index of c1, from 0 to c1.size()-1
     * @param c2_index index of c2, from 0 to c2.size()-1
     * @return this mesh index, from 0 to size()-1
     */
    std::size_t index(std::size_t c0_index, std::size_t c1_index, std::size_t c2_index) const {
        return index_f(this, c0_index, c1_index, c2_index);
    }

    /**
     * Calculate this mesh index using indexes of axis[0], axis[1] and axis[2].
     * @param indexes index of axis[0], axis[1] and axis[2]
     * @return this mesh index, from 0 to size()-1
     */
    inline std::size_t index(const Vec<3, std::size_t>& indexes) const {
        return index(indexes[0], indexes[1], indexes[2]);
    }

    /**
     * Calculate index of c0 using this mesh index.
     * @param mesh_index this mesh index, from 0 to size()-1
     * @return index of c0, from 0 to c0.size()-1
     */
    inline std::size_t index0(std::size_t mesh_index) const {
       return index0_f(this, mesh_index);
    }

    /**
     * Calculate index of c1 using this mesh index.
     * @param mesh_index this mesh index, from 0 to size()-1
     * @return index of c1, from 0 to c1.size()-1
     */
    inline std::size_t index1(std::size_t mesh_index) const {
        return index1_f(this, mesh_index);
    }

    /**
     * Calculate index of c2 using this mesh index.
     * @param mesh_index this mesh index, from 0 to size()-1
     * @return index of c2, from 0 to c2.size()-1
     */
    inline std::size_t index2(std::size_t mesh_index) const {
        return index2_f(this, mesh_index);
    }

    /**
     * Calculate index of c<axis_nr> using this mesh index.
     * @param mesh_index this mesh index, from 0 to size()-1
     * @return index of c<axis_nr>, from 0 to c<axis_nr>.size()-1
     */
    template <std::size_t axis_nr>
    inline std::size_t index_axis(std::size_t mesh_index) const {
        if (axis_nr == 0) return index0(mesh_index);
        if (axis_nr == 1) return index1(mesh_index);
        if (axis_nr == 2) return index2(mesh_index);
        assert(false);
    }

    /**
     * Calculate indexes of axes.
     * @param mesh_index this mesh index, from 0 to size()-1
     * @return index of axis[0], axis[1], and axis[2]
     */
    inline Vec<3, std::size_t> indexes(std::size_t mesh_index) const {
        return Vec<3, std::size_t>(index0(mesh_index), index1(mesh_index), index2(mesh_index));
    }

    /**
     * Calculate index of major axis using given mesh index.
     * @param mesh_index this mesh index, from 0 to size()-1
     * @return index of major axis, from 0 to majorAxis.size()-1
     */
    inline std::size_t majorIndex(std::size_t mesh_index) const {
        return mesh_index / (*minor_axis)->size() / (*medium_axis)->size();
    }

    /**
     * Calculate index of middle axis using given mesh index.
     * @param mesh_index this mesh index, from 0 to size()-1
     * @return index of middle axis, from 0 to mediumAxis.size()-1
     */
    inline std::size_t middleIndex(std::size_t mesh_index) const {
        return (mesh_index / (*minor_axis)->size()) % (*medium_axis)->size();
    }

    /**
     * Calculate index of minor axis using given mesh index.
     * @param mesh_index this mesh index, from 0 to size()-1
     * @return index of minor axis, from 0 to minorAxis.size()-1
     */
    inline std::size_t minorIndex(std::size_t mesh_index) const {
        return mesh_index % (*minor_axis)->size();
    }

    /**
     * Get number of elements (for FEM method) in the first direction.
     * @return number of elements in this mesh in the first direction (axis0 direction).
     */
    std::size_t getElementsCount0() const {
        const std::size_t s = axis[0]->size();
        return (s != 0)? s-1 : 0;
    }

    /**
     * Get number of elements (for FEM method) in the second direction.
     * @return number of elements in this mesh in the second direction (axis1 direction).
     */
    std::size_t getElementsCount1() const {
        const std::size_t s = axis[1]->size();
        return (s != 0)? s-1 : 0;
    }

    /**
     * Get number of elements (for FEM method) in the third direction.
     * @return number of elements in this mesh in the third direction (axis2 direction).
     */
    size_t getElementsCount2() const {
        const std::size_t s = axis[2]->size();
        return (s != 0)? s-1 : 0;
    }

    /**
     * Get number of elements (for FEM method).
     * @return number of elements in this mesh
     */
    size_t getElementsCount() const {
        return size_t(std::max(int(axis[0]->size())-1, 0) * std::max(int(axis[1]->size())-1, 0) * std::max(int(axis[2]->size())-1, 0));
    }

    /**
     * Convert element index to mesh index of bottom, left, front element corner.
     * @param element_index index of element, from 0 to getElementsCount()-1
     * @return mesh index
     */
    std::size_t getElementMeshLowIndex(std::size_t element_index) const {
        const std::size_t minor_size_minus_1 = (*minor_axis)->size()-1;
        const std::size_t elements_per_level = minor_size_minus_1 * ((*medium_axis)->size()-1);
        return element_index + (element_index / elements_per_level) * ((*medium_axis)->size() + minor_size_minus_1)
                            + (element_index % elements_per_level) / minor_size_minus_1;
    }

    /**
     * Convert mesh indexes of a bottom-left corner of an element to the index of this element.
     * @param axis0_index, axis1_index, axis2_index indexs of the lower corner along the axis[0, 1 and 2]
     * @return index of the element, from 0 to getElementsCount()-1
     */
    std::size_t getElementIndexFromLowIndexes(std::size_t axis0_index, std::size_t axis1_index, std::size_t axis2_index) const {
        return getElementIndexFromLowIndex(index(axis0_index, axis1_index, axis2_index));
    }

    /**
     * Check if mesh index is at the bottom, left, front corner of an element.
     * @param meshIndex mesh index
     * @return true only if @p meshIndex is at the bottom, left, front corner of an element
     */
    bool isLowIndexOfElement(std::size_t meshIndex) const {
        return index0(meshIndex) + 1 < axis[0]->size() && index1(meshIndex) + 1 < axis[1]->size() && index2(meshIndex) + 1 < axis[2]->size();
    }

    /**
     * Convert mesh index of bottom, left, front element corner to this element index.
     * @param mesh_index_of_el_bottom_left mesh index
     * @return index of the element, from 0 to getElementsCount()-1
     */
    std::size_t getElementIndexFromLowIndex(std::size_t mesh_index_of_el_bottom_left) const {
        const std::size_t verticles_per_level = (*minor_axis)->size() * (*medium_axis)->size();
        return mesh_index_of_el_bottom_left - (mesh_index_of_el_bottom_left / verticles_per_level) * ((*medium_axis)->size() + (*minor_axis)->size() - 1)
                - (mesh_index_of_el_bottom_left % verticles_per_level) / (*minor_axis)->size();
    }

    /**
     * Convert indexes of mesh axes of lower (along all axes) element corner to index of this element.
     * @param axis0_index, axis1_index, axis2_index indexes of corner
     * @return index of element, from 0 to getElementsCount()-1
     */
    std::size_t getElementIndexFromLowIndex(std::size_t axis0_index, std::size_t axis1_index, std::size_t axis2_index) const {
        return getElementIndexFromLowIndex(index(axis0_index, axis1_index, axis2_index));
    }

    /**
     * Convert element index to mesh indexes of bottom left element corner.
     * @param element_index index of element, from 0 to getElementsCount()-1
     * @return axis 0, axis 1 and axis 2 indexes of mesh,
     * you can easy calculate rest indexes of element corner adding 1 to returned coordinates
     */
    Vec<3, std::size_t> getElementMeshLowIndexes(std::size_t element_index) const {
        std::size_t bl_index = getElementMeshLowIndex(element_index);
        return Vec<3, std::size_t>(index0(bl_index), index1(bl_index), index2(bl_index));
    }

    /**
     * Calculate (using linear interpolation) value of data in point using data in points describe by this mesh.
     * @param data values of data in points describe by this mesh
     * @param point point in which value should be calculate
     * @return interpolated value in point @p point
     */
    template <typename RandomAccessContainer>
    auto interpolateLinear(const RandomAccessContainer& data, const Vec<3>& point, const InterpolationFlags& flags) const
        -> typename std::remove_reference<decltype(data[0])>::type
    {
        auto p = flags.wrap(point);

        size_t index0_lo, index0_hi;
        double back, front;
        bool invert_back, invert_front;
        prepareInterpolationForAxis(*axis[0], flags, p.c0, 0, index0_lo, index0_hi, back, front, invert_back, invert_front);

        size_t index1_lo, index1_hi;
        double left, right;
        bool invert_left, invert_right;
        prepareInterpolationForAxis(*axis[1], flags, p.c1, 1, index1_lo, index1_hi, left, right, invert_left, invert_right);

        size_t index2_lo, index2_hi;
        double bottom, top;
        bool invert_bottom, invert_top;
        prepareInterpolationForAxis(*axis[2], flags, p.c2, 2, index2_lo, index2_hi, bottom, top, invert_bottom, invert_top);

        // all indexes are in bounds
        typename std::remove_const<typename std::remove_reference<decltype(data[0])>::type>::type
            data_lll = data[index(index0_lo, index1_lo, index2_lo)],
            data_hll = data[index(index0_hi, index1_lo, index2_lo)],
            data_hhl = data[index(index0_hi, index1_hi, index2_lo)],
            data_lhl = data[index(index0_lo, index1_hi, index2_lo)],
            data_llh = data[index(index0_lo, index1_lo, index2_hi)],
            data_hlh = data[index(index0_hi, index1_lo, index2_hi)],
            data_hhh = data[index(index0_hi, index1_hi, index2_hi)],
            data_lhh = data[index(index0_lo, index1_hi, index2_hi)];

        if (invert_back)   { data_lll = flags.reflect(0, data_lll); data_llh = flags.reflect(0, data_llh); data_lhl = flags.reflect(0, data_lhl); data_lhh = flags.reflect(0, data_lhh); }
        if (invert_front)  { data_hll = flags.reflect(0, data_hll); data_llh = flags.reflect(0, data_hlh); data_lhl = flags.reflect(0, data_hhl); data_lhh = flags.reflect(0, data_hhh); }
        if (invert_left)   { data_lll = flags.reflect(1, data_lll); data_llh = flags.reflect(1, data_llh); data_hll = flags.reflect(1, data_hll); data_hlh = flags.reflect(1, data_hlh); }
        if (invert_right)  { data_lhl = flags.reflect(1, data_lhl); data_llh = flags.reflect(1, data_lhh); data_hll = flags.reflect(1, data_hhl); data_hlh = flags.reflect(1, data_hhh); }
        if (invert_bottom) { data_lll = flags.reflect(2, data_lll); data_lhl = flags.reflect(2, data_lhl); data_hll = flags.reflect(2, data_hll); data_hhl = flags.reflect(2, data_hhl); }
        if (invert_top)    { data_llh = flags.reflect(2, data_llh); data_lhl = flags.reflect(2, data_lhh); data_hll = flags.reflect(2, data_hlh); data_hhl = flags.reflect(2, data_hhh); }

        return flags.postprocess(point,
            interpolation::trilinear(back, front, left, right, bottom, top,
                                        data_lll, data_hll, data_hhl, data_lhl, data_llh, data_hlh, data_hhh, data_lhh,
                                        p.c0, p.c1, p.c2));
    }

    /**
     * Calculate (using nearest neighbor interpolation) value of data in point using data in points describe by this mesh.
     * @param data values of data in points describe by this mesh
     * @param point point in which value should be calculate
     * @return interpolated value in point @p point
     */
    template <typename RandomAccessContainer>
    auto interpolateNearestNeighbor(const RandomAccessContainer& data, Vec<3> point, const InterpolationFlags& flags) const
        -> typename std::remove_reference<decltype(data[0])>::type {
        auto p = flags.wrap(point);
        prepareNearestNeighborInterpolationForAxis(*axis[0], flags, p.c0, 0);
        prepareNearestNeighborInterpolationForAxis(*axis[1], flags, p.c1, 1);
        prepareNearestNeighborInterpolationForAxis(*axis[2], flags, p.c2, 2);
        return flags.postprocess(point, data[this->index(axis[0]->findNearestIndex(p.c0), axis[1]->findNearestIndex(p.c1), axis[2]->findNearestIndex(p.c2))]);
    }

protected:
    bool hasSameNodes(const MeshD<3> &to_compare) const override {
        if (const RectilinearMesh3D* c = dynamic_cast<const RectilinearMesh3D*>(&to_compare))
            return *this == *c;  // this will call == operator from RectilinearMesh3D
        return RectangularMeshBase3D::hasSameNodes(to_compare);
    }

private:

    /**
     * Iterator over plane CHANGE_DIR_SLOWER, CHANGE_DIR_FASTER (an index if the remain coordinate is constant)
     *
     * Each increment() changes CHANGE_DIR_FASTER index.
     *
     * Common code for: left, right, bottom, top, front, back boundaries.
     */
    template <int CHANGE_DIR_SLOWER, int CHANGE_DIR_FASTER>
    struct BoundaryIteratorImpl: public plask::BoundaryNodeSetImpl::IteratorImpl {

        const RectilinearMesh3D &mesh;

        /// current indexes
        Vec<3, std::size_t> index;

        /// past the last index of change direction
        const std::size_t indexFasterBegin, indexFasterEnd, indexSlowerEnd;

    public:
        BoundaryIteratorImpl(const RectilinearMesh3D& mesh, Vec<3, std::size_t> index, std::size_t indexSlowerEnd, std::size_t indexFasterEnd)
            : mesh(mesh), index(index), indexFasterBegin(index[CHANGE_DIR_FASTER]), indexFasterEnd(indexFasterEnd), indexSlowerEnd(indexSlowerEnd)
        {
        }

        void increment() override {
            ++index[CHANGE_DIR_FASTER];
            if (index[CHANGE_DIR_FASTER] == indexFasterEnd) {
                index[CHANGE_DIR_FASTER] = indexFasterBegin;
                ++index[CHANGE_DIR_SLOWER];
            }
        }

        bool equal(const plask::BoundaryNodeSetImpl::IteratorImpl& other) const override {
            return index == static_cast<const BoundaryIteratorImpl&>(other).index;
        }

        std::size_t dereference() const override {
            return mesh.index(index);
        }

        std::unique_ptr<plask::BoundaryNodeSetImpl::IteratorImpl> clone() const override {
            return std::unique_ptr<plask::BoundaryNodeSetImpl::IteratorImpl>(new BoundaryIteratorImpl<CHANGE_DIR_SLOWER, CHANGE_DIR_FASTER>(*this));
        }

    };

    template <int CHANGE_DIR_SLOWER, int CHANGE_DIR_FASTER>
    struct BoundaryNodeSetImpl: public BoundaryNodeSetWithMeshImpl<RectilinearMesh3D> {

        static constexpr int FIXED_DIR = 3 - CHANGE_DIR_SLOWER - CHANGE_DIR_FASTER;

        using typename BoundaryNodeSetWithMeshImpl<RectilinearMesh3D>::const_iterator;

        /// fixed index
        std::size_t level;

        /// past the last index of change directions
        std::size_t indexFasterEnd, indexSlowerEnd;

        BoundaryNodeSetImpl(const RectilinearMesh3D& mesh, std::size_t level)
            : BoundaryNodeSetWithMeshImpl<RectilinearMesh3D>(mesh), level(level) {}

        bool contains(std::size_t mesh_index) const override {
            return mesh_index < this->mesh.size() && this->mesh.template index_axis<FIXED_DIR>(mesh_index) == level;
        }

        const_iterator begin() const override {
            Vec<3, std::size_t> index_begin(0, 0, 0);
            index_begin[FIXED_DIR] = level;
            return Iterator(new BoundaryIteratorImpl<CHANGE_DIR_SLOWER, CHANGE_DIR_FASTER>(this->mesh, index_begin,
                                                                                           this->mesh.axis[CHANGE_DIR_SLOWER]->size(),
                                                                                           this->mesh.axis[CHANGE_DIR_FASTER]->size()));
        }

        const_iterator end() const override {
            Vec<3, std::size_t> index_end;
            index_end[CHANGE_DIR_SLOWER] = this->mesh.axis[CHANGE_DIR_SLOWER]->size();
            index_end[FIXED_DIR] = level;
            index_end[CHANGE_DIR_FASTER] = 0;
            return Iterator(new BoundaryIteratorImpl<CHANGE_DIR_SLOWER, CHANGE_DIR_FASTER>(this->mesh, index_end,
                                                                                           this->mesh.axis[CHANGE_DIR_SLOWER]->size(),
                                                                                           this->mesh.axis[CHANGE_DIR_FASTER]->size()));
        }

        std::size_t size() const override {
            return this->mesh.axis[CHANGE_DIR_FASTER]->size() * this->mesh.axis[CHANGE_DIR_SLOWER]->size();
        }

        bool empty() const override {
            return this->mesh.axis[CHANGE_DIR_FASTER]->empty() || this->mesh.axis[CHANGE_DIR_SLOWER]->empty();
        }
    };

    template <int CHANGE_DIR_SLOWER, int CHANGE_DIR_FASTER>
    struct BoundaryNodeSetRangeImpl: public BoundaryNodeSetWithMeshImpl<RectilinearMesh3D> {

        static constexpr int FIXED_DIR = 3 - CHANGE_DIR_SLOWER - CHANGE_DIR_FASTER;

        typedef typename BoundaryNodeSetImpl::Iterator Iterator;

        /// first index
        Vec<3, std::size_t> index;

        /// past the last index of change directions
        std::size_t indexFasterEnd, indexSlowerEnd;

        BoundaryNodeSetRangeImpl(const RectilinearMesh3D& mesh, Vec<3, std::size_t> index, std::size_t indexSlowerEnd, std::size_t indexFasterEnd)
            : BoundaryNodeSetWithMeshImpl<RectilinearMesh3D>(mesh), index(index), indexFasterEnd(indexFasterEnd), indexSlowerEnd(indexSlowerEnd) {}

        BoundaryNodeSetRangeImpl(const RectilinearMesh3D& mesh, std::size_t index0, std::size_t index1, std::size_t index2, std::size_t indexSlowerEnd, std::size_t indexFasterEnd)
            : BoundaryNodeSetWithMeshImpl<RectilinearMesh3D>(mesh), index(index0, index1, index2), indexFasterEnd(indexFasterEnd), indexSlowerEnd(indexSlowerEnd) {}

        bool contains(std::size_t mesh_index) const override {
            if (mesh_index >= this->mesh.size()) return false;
            Vec<3, std::size_t> mesh_indexes = this->mesh.indexes(mesh_index);
            return mesh_indexes[FIXED_DIR] == index[FIXED_DIR] &&
                   (index[CHANGE_DIR_FASTER] <= mesh_indexes[CHANGE_DIR_FASTER] && mesh_indexes[CHANGE_DIR_FASTER] < indexFasterEnd) &&
                   (index[CHANGE_DIR_SLOWER] <= mesh_indexes[CHANGE_DIR_SLOWER] && mesh_indexes[CHANGE_DIR_SLOWER] < indexSlowerEnd);
        }

        const_iterator begin() const override {
            return Iterator(new BoundaryIteratorImpl<CHANGE_DIR_SLOWER, CHANGE_DIR_FASTER>(this->mesh, index, indexSlowerEnd, indexFasterEnd));
        }

        const_iterator end() const override {
            Vec<3, std::size_t> index_end = index;
            index_end[CHANGE_DIR_SLOWER] = indexSlowerEnd;
            return Iterator(new BoundaryIteratorImpl<CHANGE_DIR_SLOWER, CHANGE_DIR_FASTER>(this->mesh, index_end, indexSlowerEnd, indexFasterEnd));
        }

        std::size_t size() const override {
            return (indexFasterEnd - index[CHANGE_DIR_FASTER]) * (indexSlowerEnd - index[CHANGE_DIR_SLOWER]);
        }

        bool empty() const override {
            return indexFasterEnd == index[CHANGE_DIR_FASTER] || indexSlowerEnd == index[CHANGE_DIR_SLOWER];
        }
    };

  public:

  BoundaryNodeSet createIndex0BoundaryAtLine(std::size_t line_nr_axis0) const override;

  BoundaryNodeSet createBackBoundary() const override;

  BoundaryNodeSet createFrontBoundary() const override;

  BoundaryNodeSet createIndex1BoundaryAtLine(std::size_t line_nr_axis1) const override;

  BoundaryNodeSet createLeftBoundary() const override;

  BoundaryNodeSet createRightBoundary() const override;

  BoundaryNodeSet createIndex2BoundaryAtLine(std::size_t line_nr_axis2) const override;

  BoundaryNodeSet createBottomBoundary() const override;

  BoundaryNodeSet createTopBoundary() const override;

  BoundaryNodeSet createIndex0BoundaryAtLine(std::size_t line_nr_axis0,
                                             std::size_t index1Begin, std::size_t index1End,
                                             std::size_t index2Begin, std::size_t index2End) const override;

  BoundaryNodeSet createBackOfBoundary(const Box3D& box) const override;

  BoundaryNodeSet createFrontOfBoundary(const Box3D& box) const override;

  BoundaryNodeSet createIndex1BoundaryAtLine(std::size_t line_nr_axis1,
                                             std::size_t index0Begin, std::size_t index0End,
                                             std::size_t index2Begin, std::size_t index2End) const override;

  BoundaryNodeSet createLeftOfBoundary(const Box3D& box) const override;

  BoundaryNodeSet createRightOfBoundary(const Box3D& box) const override;

  BoundaryNodeSet createIndex2BoundaryAtLine(std::size_t line_nr_axis2,
                                             std::size_t index0Begin, std::size_t index0End,
                                             std::size_t index1Begin, std::size_t index1End) const override;

  BoundaryNodeSet createBottomOfBoundary(const Box3D& box) const override;

  BoundaryNodeSet createTopOfBoundary(const Box3D& box) const override;
};


}   // namespace plask

#endif // PLASK__RECTILINEAR3D_H
